/*
 * packetmodeller/model.c
 *
 * Copyright (C) 2009 Reto Gantenbein
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include "message.h"
#include "model.h"
#include "packet_modeller.h"
#include "pm_print.h"

/** Declare local wait indicator */
int wait;

/** Get vif information for model */
int
get_model_vif_info(model_vif_info *model_vif, const pm_vif_info *pm_vif)
{
    strncpy(model_vif->vif_name, pm_vif->vif_data->vif_name, IFNAMSIZ + 1);
    strncpy(model_vif->vif_id, pm_vif->vif_id, sizeof(host_id));
    memcpy(model_vif->vif_mac, pm_vif->vif_mac, ETH_ALEN);
    model_vif->freq = pm_vif->vif_data->vif_config.b.freq;
    return 0;
}

/** Clears wait status */
void
clear_wait()
{
    wait = 0;
}

/** Register the host at the WlanModel */
int
pm_register()
{
    struct register_msg *reg_msg;
    model_vif_info *model_vif;
    char reg_buffer[REGISTER_MSG_SIZE];
    char ack_buffer[WLAN_MSG_SIZE];
    uint16_t message_id = 0;
    int recv_bytes = 0;
    uint8_t success = 0;

    reg_msg = (struct register_msg *) reg_buffer;
    (*reg_msg).header.type = REGISTER_MSG;
    strncpy((char *) &(*reg_msg).header.id, (char *) node_id, sizeof(host_id));
    strncpy((char *) &(*reg_msg).name, host_name, HOST_NAME_MAX);
    memmove(&(*reg_msg).ip, &node_ip, sizeof(struct in_addr));
    reg_msg->port = node_port;
    model_vif = &(reg_msg->vif_info);
    get_model_vif_info(model_vif, &vif_list[0]);
    signal(SIGALRM, clear_wait);

    while ((success == 0) && (message_id < 10))
    {
        (*reg_msg).msg_id = message_id;
        wait = 1;
        debug_msg((char *) reg_msg);

        // send register message to model
        if (sendto(net_socket, reg_msg, REGISTER_MSG_SIZE, 0, \
            (struct sockaddr *) &model_addr, sizeof(struct sockaddr_in)) < 0)
        {
            pm_error("Couldn't send register message (%s)\n", \
            strerror(errno));
        }
        pm_debug("register_node(): [net_socket] %i bytes sent (%s)\n", \
            REGISTER_MSG_SIZE, strerror(errno));

        // wait 5s for ack message else resend register message
        alarm(5);
        while (wait == 1)
        {
            if ((recv_bytes = recv(net_socket, ack_buffer, \
                sizeof(ack_buffer), MSG_DONTWAIT)) > 0)
            {
                pm_debug("register_node(): [net_socket] %i bytes read (%s)\n", \
                    recv_bytes, strerror(errno));
                debug_msg(ack_buffer);
                if (isAcknowledgeMessage((struct msg_header *) ack_buffer) && \
                    (((struct ack_msg *) ack_buffer)->msg_id <= message_id))
                {
                    alarm(0);  // reset alarm
                    wait = 0;
                    success = 1;
                }
            } else {
                errno = 0;
            }
        }
        message_id++;
    }
    if (success == 0)
    {
        pm_error("Couldn't register node at WlanModel!\n");
        return -1;
    }
    pm_log("Host successfully registered at WlanModel\n");

    return 0;
}

/** Sends config_msg to update vif parameter at WlanModel */
int
update_model(int vif_index, int property)  //TODO: handle correct vif_index
{
    struct config_msg *notify_msg;

    notify_msg = malloc(sizeof(struct config_msg));
    (*notify_msg).header.type = CONFIG_MSG;
    strncpy((char *) &(*notify_msg).header.id, \
        (char *) node_id, sizeof(host_id));
    switch (property)
    {
        case SIOCSIWFREQ:
            notify_msg->property = property;
            notify_msg->value.freq = vif_list[0].vif_data->vif_config.b.freq;
            break;
        default:
            pm_debug("update_model(): property %x not found!\n", property);
            break;
    }
    debug_msg((char *) notify_msg);

    // send data to model
    if (sendto(net_socket, notify_msg, CONFIG_MSG_SIZE, 0, \
        (struct sockaddr *) &model_addr, sizeof(struct sockaddr_in)) < 0)
    {
        pm_error("Sending data failed (%s)\n", strerror(errno));
    }
    pm_debug("update_model(): [net_socket] %i Bytes sent (%s)\n", \
        CONFIG_MSG_SIZE, strerror(errno));
    free(notify_msg);

    return 0;
}

/** Handles model updates on vif configuration changes */
void *
vif_handler()
{
    pm_msg msg_buffer;
    int msgq_id = -1;

    pm_debug("vif_handler(): Starting thread\n");

    // connecting to message queue with vifd
    if ((msgq_id = msgget(IPC_KEY, 0600)) < 0)
    {
        perror("msgget");
        return (void *) -1;
    }
    while (1)
    {
        memset(&msg_buffer, 0, sizeof(pm_msg));
        if (msgrcv(msgq_id, &msg_buffer, sizeof(pm_msg), PM_UPDATE, 0) > 0)
        {
            update_model(msg_buffer.vif_index, msg_buffer.property);
        } else {
            perror("msgrcv");       // raise error, but continue execution
        }

    }

    return (void *) 0;
}
