/*
 * packetmodeller/netdevice.c
 *
 * Copyright (C) 2008-2009 Reto Gantenbein
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include "netdevice.h"
#include "pm_print.h"

/** Fill interface request with information of given device */
int
get_ifr_by_dev(struct ifreq *ifr_ext, const char *dev_name)
{
    int fd, i;
    int ifrs_size = 10;
    struct ifconf ifc;
    struct ifreq *ifr_p;
    struct ifreq ifrs[ifrs_size];

    if ((fd = socket(PF_INET, SOCK_DGRAM, 0)) == -1)
    {
        pm_error("Couldn't open socket %s (%s)\n", dev_name, strerror(errno));
        return -1;
    }
    memset(&ifrs, 0, sizeof(ifrs));
    ifc.ifc_len = sizeof(ifrs);
    ifc.ifc_req = ifrs;
    if (ioctl(fd, SIOCGIFCONF, (void *) &ifc) < 0)
    {
        close(fd);
        pm_debug("%s:%i - Couldn't fetch device configuration (%s)\n", \
            __FILE__, __LINE__, strerror(errno));
        return -1;
    }
    if (ifc.ifc_len == sizeof(ifrs))
    {
        close(fd);
        pm_debug("%s:%i - Couldn't fetch all device information \
            (ifc_len too small)\n", __FILE__, __LINE__);
        return -1;
    }
    ifrs_size = ifc.ifc_len / sizeof(struct ifreq);
    for (i = 0; i < ifrs_size; i++)
    {
        ifr_p = ((struct ifreq*) ifc.ifc_req) + i;
        if (strcmp(dev_name, (*ifr_p).ifr_name) == 0)
        {
            memcpy(ifr_ext, ifr_p, sizeof(struct ifreq));
            return 0;
        }
    }
    return -1;
}

/** Return the mac address of the given interface */
int get_mac_by_dev(unsigned char *mac, const char *dev_name)
{
    int fd;
    struct ifreq ifr;

    strncpy((char *) &(ifr.ifr_name), (char *) dev_name, IFNAMSIZ + 1);
    if ((fd = socket(PF_INET, SOCK_DGRAM, 0)) == -1)
    {
        pm_error("Couldn't open socket on %s (%s)\n", dev_name, strerror(errno));
        return -1;
    }
    if (ioctl(fd, SIOCGIFHWADDR, &ifr) != 0)
    {
        pm_error("Couldn't get MAC address of interface %s (%s)\n", \
            dev_name, strerror(errno));
        return -1;
    }
    memcpy(mac, ifr.ifr_hwaddr.sa_data, 6);
    pm_debug("%s: %2.2x:%2.2x:%2.2x:%2.2x:%2.2x:%2.2x\n", \
        dev_name, mac[0], mac[1], mac[2], mac[3], mac[4], mac[5]);

    return 0;
}

/** Return ip address to a given interface */
int get_ip_by_dev(struct in_addr *addr, const char *dev_name)
{
    struct ifreq ifr;

    if (get_ifr_by_dev(&ifr, dev_name) == 0)
    {
        if (addr != NULL) {
            // if we only check for valid interface we don't need addr struct
            struct sockaddr_in *sockaddr =
                (struct sockaddr_in *) &(ifr.ifr_addr);
            memcpy(addr, (struct in_addr *) &(*sockaddr).sin_addr, \
                sizeof(struct in_addr));
        }
        return 0;
    } else {
        return -1;
    }
}

/** Initialize address struct with data from given interface and port */
int
get_addr(struct sockaddr_in *addr, const char *dev_name, const uint16_t port)
{
    get_ip_by_dev(&(*addr).sin_addr, dev_name);
    addr->sin_family = AF_INET;
    addr->sin_port = htons(port);
    memset(addr->sin_zero, '\0', sizeof(addr->sin_zero));
    return 0;
}

/** Create UDP socket and bind to given interface */
int
alloc_net_socket(const char *dev_name, const uint16_t port)
{
    int fd = 0;
    struct sockaddr_in *local_addr;

    // use SOCK_DGRAM (=> UDP)
    if ((fd = socket(PF_INET, SOCK_DGRAM, 0)) == -1)
    {
        pm_error("Cannot open socket on %s (%s)\n", dev_name, strerror(errno));
        return EXIT_FAILURE;
    }
    local_addr = calloc(1, sizeof(struct sockaddr_in));
    get_addr(local_addr, dev_name, port);
    if (bind(fd, (struct sockaddr *) local_addr, \
        (socklen_t) sizeof(struct sockaddr)) == -1)
    {
        pm_error("Couldn't bind to local interface %s (%s)\n", \
            dev_name, strerror(errno));
    }
    free(local_addr);
    return fd;
}

/** Return character device file descriptor for TAP device */
int alloc_tap_socket(const char * dev_name)
{
    int fd = 0, err = 0;
    struct ifreq ifr;

    fd = open("/dev/net/tun", O_RDWR);
    if (fd < 0)
    {
        pm_error("Couldn't open /dev/net/tun (%s)\n", strerror(errno));
        return -1;
    }
    memset(&ifr, 0, sizeof(ifr));
    ifr.ifr_flags = IFF_TAP;    // set flag for TAP device
    ifr.ifr_flags = IFF_NO_PI;  // we don't need packet info (?)
    if (*dev_name)
    {
       memcpy(ifr.ifr_name, dev_name, IFNAMSIZ);
    }
    err = ioctl(fd, TUNSETIFF, (void *) &ifr) && ioctl(fd, TUNSETNOCSUM, 1);
    if (err < 0)
    {
        close(fd);
        pm_debug("Warning: ioctl on %s failed (%s)\n", \
            dev_name, strerror(errno));
        return -1;
    }

    return fd;
}

/** Check if it's available device name */
int
validate_dev_name(const char * dev_name)
{
    char iface[IFNAMSIZ];
    extern char node_pif[IFNAMSIZ];

    int if_len = strlen(dev_name);
    strncpy(iface, dev_name, if_len);
    iface[if_len] = '\0';
    if (get_ip_by_dev(NULL, iface) == -1)
    {
        pm_error("Invalid interface %s\n", iface);
        return -1;
    }
    strncpy(node_pif, iface, IFNAMSIZ);

    return 0;
}

/** Checks char array for valid IP address (0.0.0.0 to 255.255.255.255) */
int
validate_ip(const char *ip)
{
    int err = 0;
    unsigned b1, b2, b3, b4;

    if (sscanf(ip, "%3u.%3u.%3u.%3u", &b1, &b2, &b3, &b4) != 4)
        err = 1;
    if (b1 > 255 || b2 > 255 || b3 > 255 || b4 > 255)
        err = 1;
    if (err)
        printf("Error: Invalid model-ip address given\n");

    return err;
}

/** Read a port from char array, check for validity and return it in
 *  network byte order */
uint16_t
validate_port(const char *cport)
{
    uint16_t iport = atoi(cport);
    if (iport > 0) // don't need to check upper limit becaues of uint16_t range
    {
        return htons(iport);
    } else {
        printf("Error: Invalid port given\n");
        return 0;
    }
}

/** Checks IP:Port pair for validity and fills into sockaddr_in struct */
int
validate_model_addr(struct sockaddr_in *addr, const char *chost)
{
    int ip_char;
    const char delimiter = ':';

    for (ip_char = 0; ip_char < 16; ip_char++)
    {
        if (*(chost + ip_char) == delimiter)
        {
            char ip_array[ip_char + 1];
            memcpy(ip_array, chost, ip_char);
            ip_array[ip_char] = '\0';
            if (validate_ip(ip_array) != 0)
                return -1;
            inet_aton(ip_array, &((*addr).sin_addr));
            if (((*addr).sin_port = validate_port(chost + ip_char + 1)) == 0)
                return -1;
            (*addr).sin_family = AF_INET;
            return 0;
        }
    }
    printf("Error: Invalid format model-ip:model-port given\n");

    return -1;
}
