/*
 * vifd/vifd.c
 *
 * Copyright (C) 2008-2009 Reto Gantenbein
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

/**
 * VifD - Virtual Interface Daemon
 *
 * Usage: vifd <device_name> [<device_name>]
 *
 * Create a virtual wireless interface via TUN/TAP driver. The initial device
 * configuration should be consistent with an unconfigured Atheros WLAN device.
 *
 * It communicates via simulated ioctls via message queue with a patched version
 * of the wireless-tools, to set and get the virtual device wireless parameters.
 */

#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <syslog.h>

#include "libvif.h"
#include "vifd_ipc.h"
#include "vifd_print.h"


/** VIFD VALUES ***************************************************************/

/** Size of shared memory address space */
#define SHARED_MEM_SIZE     sizeof(vif_private_args)


/** VIFD VARIABLES ************************************************************/

/** String of program name */
char *exec_name;
int debug;

/** Message queue identifiers */
int iw_msgq_id;
int pm_msgq_id;

/** Message queue identifiers */
int iw_shm_id;
int pm_shm_id;


/** PRIVATE FUNCTIONS (DON'T CALL FROM OUTSIDE) *******************************/

/**
 * Print program usage to stdout
 */
void
vifd_usage()
{
    printf("\nUsage:\t%s VIFNAME [VIFNAME] [-d]\n", exec_name);
    printf("\tCreate virtual wireless interfaces (-d for debug)\n");
    exit(EXIT_FAILURE);
}

/** Stub for print_debug prototype */
int
ignore_debug(const char* text, ...)
{
    return 0;
}

/** Print debug message with variable argument list */
int
print_debug(const char *format, ...)
{
    va_list arg;

    va_start(arg, format);
    vifd_debug_args(format, arg);
    va_end(arg);

    return 0;
}

/** Stub for print_debug_args prototype */
void
ignore_debug_args(const char *format, va_list arg)
{
}

/** Print debug message with given argument list */
void
print_debug_args(const char *format, va_list arg)
{
    FILE *stdout = fopen("/dev/stdout", "a");
    fprintf(stdout, "DEBUG: ");
    vfprintf(stdout, format, arg);
    fclose(stdout);
}

/** Print Error message */
void
vifd_error(const char *format, ...)
{
    va_list arg;

    FILE *stderr = fopen("/dev/stderr", "a");
    fprintf(stderr, "Error: ");
    va_start(arg, format);
    vfprintf(stderr, format, arg);
    va_end(arg);
    fclose(stderr);
}

/** Send message to syslog */
void
vifd_log(const char *format, ...)
{
    va_list arg;

    openlog(exec_name, LOG_PID, LOG_SYSLOG);
    va_start(arg, format);
    vsyslog(LOG_INFO, format, arg);
    vifd_debug_args(format, arg);
    va_end(arg);
    closelog();
}

/**
 * Check for valid arguments
 */
void
check_args(int num_args, char *args[])
{
    int i, opt;

    // check for debugging option
    while ((opt = getopt(num_args, args, "d")) != -1)
    {
        if (opt == 'd')
        {
            debug = 1;
        } else {
            debug = 0;
        }
    }
    if (debug == 1)
    {
        vifd_debug = print_debug;
        vifd_debug_args = print_debug_args;
    } else {
        vifd_debug = ignore_debug;
        vifd_debug_args = ignore_debug_args;
    }

    // check for vif arguments
    if ((num_args == 1) || ((debug == 1) && (num_args == 2)))
    {
        vifd_error("Missing argument\n");
        vifd_usage();
    } else {
        if (num_args > MAX_NUM_VIF + 1)
        {
            vifd_error("Only %i virtual interfaces supported\n", MAX_NUM_VIF);
            vifd_usage();
        }
    }
    for (i = 1; i < num_args; i++)
    {
        if (strlen(args[i]) > IF_NAMESIZE)
        {
            vifd_error("Interface name too long\n");
            vifd_usage();
        }
    }
    vifd_log("Starting %s (v0.1.0)\n", exec_name);
}

/**
 * check executable
 */
int
check_exec(char *exec, int len)
{
    char check_cmd[6 + len];

    sprintf(check_cmd, "test -x %s", exec);
    int rtrn = system(check_cmd);
    if (rtrn != 0)
    {
        vifd_error("%s not found or not an executable\n", exec);
        exit(EXIT_FAILURE);
    }

    return rtrn;
}

/**
 * Check for virtual interface
 */
int
check_vif(char *vif_name)
{
    char cmd[60 + IF_NAMESIZE];

    check_exec(IP_PATH, 20);
    sprintf(cmd, "%s link show dev %s >> /dev/null 2>&1", IP_PATH, vif_name);

    return system(cmd);
}

/**
 * Create tun-tap interface
 */
void
create_tap(char *vif_name)
{
    char cmd[40 + IF_NAMESIZE];

    check_exec(TUNCTL_PATH, 20);
    sprintf(cmd, "%s -t %s >> /dev/null 2>&1", TUNCTL_PATH, vif_name);
    int rtrn = system(cmd);
    if  (rtrn != 0)
    {
        vifd_error("Couldn't create virtual device %s\n", vif_name);
        exit(EXIT_FAILURE);
    }
    vifd_log("Virtual wifi interface %s created\n", vif_name);
}

/**
 * Remove tun-tap interface
 */
void
remove_tap(char *vif_name)
{
    char cmd[40 + IF_NAMESIZE];

    check_exec(TUNCTL_PATH, 20);
    sprintf(cmd, "%s -d %s >> /dev/null 2>&1", TUNCTL_PATH, vif_name);
    int rtrn = system(cmd);
    if  (rtrn != 0)
    {
        vifd_error("Couldn't remove virtual device %s\n", vif_name);
    }
}

/**
 * Shutdown vifd process and remove virtual interfaces
 */
static void
vifd_shutdown(int signo)
{
    int i, rtrn;
    char * vif_name;
    char vif_enum[num_vif * (IFNAMSIZ + 3)];
    char *vifp = vif_enum;

    msgq_delete(iw_msgq_id);
    msgq_delete(pm_msgq_id);
    shmem_delete(iw_shm_id);
    shmem_delete(pm_shm_id);
    for (i = 0; i < num_vif; i++)
    {
        vif_name = vif_list[i]->vif_name;
        if ((rtrn = check_vif(vif_name)) == 0)
        {
            remove_tap(vif_name);
            sprintf(vifp, " '%s'", vif_name);
            vifp += (3 + strlen(vif_name));
        }
    }
    vifd_log("Shutdown vifd\n");
    if (num_vif > 0)
        vifd_log("Removing device(s):%s\n", vif_enum);

    exit(EXIT_SUCCESS);
}

/**
 * Poll message queue for ioctl calls
 */
void
vifd_daemon(void)
{
    struct iwreq_msg msg_buffer;

    if (iw_msgq_id < 0)
    {
        perror("invalid message queue");
        exit(EXIT_FAILURE);
    }
    memset(&msg_buffer, 0, sizeof(struct iwreq_msg));
    while (1)
    {
        msgq_receive(iw_msgq_id, &msg_buffer, IW_MSG_SIZE, IW_IOCTL);
        handle_iwreq(&msg_buffer);
    }
}

/**
 * MAIN
 */
int
main(int argc, char *argv[])
{
    int i;
    char *vif_name;

    if (strrchr(argv[0], '/') == NULL)
    {
        exec_name = argv[0];
    } else {
        exec_name = strrchr(argv[0], '/') + 1;
    }
    check_args(argc, argv);
    num_vif = 0;

    // registering signal handler
    signal(SIGINT, vifd_shutdown);
    signal(SIGABRT, vifd_shutdown);
    signal(SIGTERM, vifd_shutdown);

    // reserve shared memory for vif configurations
    pm_msgq_id = msgq_initialize(PM_IPC_KEY);
    pm_shm_id = shmem_initialize(PM_IPC_KEY, \
        MAX_NUM_VIF * sizeof(struct vif_info));
    for (i = 1; i < argc; i++)
    {
        if (strcmp(argv[i], "-d") != 0)
        {
            vif_name = argv[i];
            if (check_vif(vif_name) != 0)
            {
                create_tap(vif_name);
            } else {
                vifd_error("Device '%s' already exists\n", vif_name);
                vifd_shutdown(0);
            }
            create_vif_config(vif_name, num_vif);
        }
    }
    vifd_debug("main(): num_vif = %i\n", num_vif);
    iw_msgq_id = msgq_initialize(IW_IPC_KEY);
    iw_shm_id = shmem_initialize(IW_IPC_KEY, SHARED_MEM_SIZE);
    vifd_debug("main(): message queues initialized (iw_id: %i, pm_id: %i)\n", \
        iw_msgq_id, pm_msgq_id);
    vifd_debug("main(): shared memory initialized (iw_id: %i, pm_id: %i)\n", \
        iw_shm_id, pm_shm_id);

    // fork to execute in background
    if (debug != 1)
    {
        i = fork();
        if (i < 0)
        {
            vifd_error("Couldn't fork process to background\n");
            vifd_shutdown(0);
        }
        if (i > 0)
        {
            exit(0); //parent exit
        }
    }

    // run daemon for interface changes
    vifd_daemon();

    return EXIT_SUCCESS;
}
